from . import txt2img,config
from nonebot.plugin import PluginMetadata
from .version import __version__,url
from .others.deepdanbooru import deepdanbooru
__plugin_meta__ = PluginMetadata(
    name="AI绘图",
    description="调用novelai进行二次元AI绘图",
    usage=f"基础用法:\n.aidraw[指令] [空格] loli,[参数]\n示例:.aidraw loli,cute,kawaii,\n项目地址:{url}",
    extra={
        "version":__version__
    }
)


from importlib.metadata import version
import asyncio
from .utils import check_last_version,sendtosuperuser
from nonebot.log import logger
import time

package="nonebot-plugin-novelai"
url="https://github.com/Mutsukibot/tree/nonebot-plugin-novelai"
try:
    __version__=version(package)
except:
    __version__="0.4.1"

lastcheck:float=0
async def check_update(isinit=True):
    global lastcheck
    if time.time()-lastcheck>80000:
        update=await check_last_version(package,__version__)
        if update:
            message=f"检测到新版本{update},当前版本{__version__},请使用pip install --upgrade {package}命令升级,查看项目获取更多信息:{url}"
            logger.info(message)
            if isinit:
                await sendtosuperuser("novelai插件"+message)
        else:
            logger.info(f"检查版本完成，当前为最新版本")
        lastcheck=time.time()
asyncio.run(check_update(False))


from pydantic import BaseSettings, validator
from nonebot import get_driver
from nonebot.log import logger
from pydantic.fields import ModelField
import asyncio

class Config(BaseSettings):
    novelai_token: str = ""
    novelai_tag: str = ""
    novelai_cd: int = 60
    novelai_limit: bool = True
    novelai_save_pic:bool = True
    novelai_api_domain: str = "https://api.novelai.net/"
    novelai_site_domain: str = "https://novelai.net/"
    novelai_mode:str = "novelai"
    novelai_paid:int = 0
    novelai_ban:list[int]=[]
    novelai_h:bool = False
    novelai_oncemax:int = 3
    bing_key:str=None

    @validator("novelai_cd","novelai_oncemax")
    def non_negative(cls, v: int, field: ModelField):
        if v < 1:
            return field.default
        return v
    @validator("novelai_paid")
    def paid(cls, v: int, field: ModelField):
        if v < 0:
            return field.default
        elif v > 2:
            return field.default
        return v

    def check_mode(cls):
        mode=cls.novelai_mode
        match mode:
            case "novelai":
                if cls.novelai_token:
                    cls.set_novelai()
                else:
                    logger.error(f"使用novelai模式但未检测到novelai token，已切换至public_naifu模式，该模式下运行不稳定")
                    asyncio.run(cls.setup_naifu())
                    cls.mode="public_naifu"
            case "public_naifu":
                asyncio.run(cls.setup_naifu())
                logger.info(f"正在使用public_naifu模式，该模式下运行不稳定")
            case "public_webui":
                asyncio.run(cls.setup_webui())
                logger.info(f"正在使用webui模式，该模式下运行不稳定")
            case "mix":
                logger.info(f"正在使用mix(public_naifu+public_naifu)模式，该模式下运行不稳定")
            case "naifu":
                if cls.novelai_api_domain == "https://api.novelai.net/" or cls.novelai_site_domain == "https://novelai.net/":
                    logger.error(f"请配置个人服务器api和site，已切换至public_naifu模式，该模式下运行不稳定")
                    cls.mode="public_naifu"
            case "webui":
                if cls.novelai_api_domain == "https://api.novelai.net/" or cls.novelai_site_domain == "https://novelai.net/":
                    logger.error(f"请配置个人服务器api和site，已切换至public_webui模式，该模式下运行不稳定")
                    cls.mode="public_webui"
            case _:
                if cls.novelai_token:
                    logger.error(f"请配置正确的运行模式，检测到token，自动切换至novelai模式")
                    cls.set_novelai()
                    cls.mode="novelai"
                else:
                    logger.error(f"请配置正确的运行模式，自动切换至public_naifu模式，该模式下运行不稳定")
                    cls.mode="public_naifu"

    def set_novelai(cls):
        cls.novelai_api_domain: str = "https://api.novelai.net/"
        cls.novelai_site_domain: str = "https://novelai.net/"
    async def setup_naifu():
        pass
    def get_api():
        pass

    def set_enable(cls,enable,groupid):
        if groupid in cls.novelai_ban:
            if enable:
                return f"aidraw已经处于启动状态"
            else:
                cls.novelai_ban.pop(groupid)
                return f"aidraw已关闭"
        else:
            if enable:
                cls.novelai_ban.append(groupid)
                return f"aidraw开始运行"
            else:
                return f"aidraw已经处于关闭状态"
    class Config:
        extra = "ignore"

config = Config(**get_driver().config.dict())
config.check_mode()
logger.debug(f"加载config完成" + str(config))



from dataclasses import dataclass
from io import BytesIO
from PIL import Image
import base64
from .config import config


@dataclass()
class TXT2IMG():
    width: int
    height: int
    image:None=None


@dataclass()
class IMG2IMG(TXT2IMG):
    image: str

    def __init__(self, img_b64):
        tmpfile = BytesIO(img_b64)
        image = Image.open(tmpfile)
        width, height = image.size
        if width >= height:
            self.width = round(width/height*8)*64
            self.height = 512
        else:
            self.height = round(height/width*8)*64
            self.width = 512
        self.image = str(base64.b64encode(img_b64), "utf-8")


@dataclass
class FIFO_TXT():
    user_id: int
    tags: str
    seed: int
    data: TXT2IMG
    group_id: int = 0
    strength: float = 0.7
    count: int = 1
    cost:None=None



@dataclass
class FIFO_IMG(FIFO_TXT):
    data: list[IMG2IMG]
    cost: int = 0
    def __post_init__(self):
        if config.novelai_paid == 1:
            anlas = 0
            for i in self.data:
                width = i.width
                height = i.height
                anlas += round(width*height*self.strength*self.count/45875)
            self.cost = anlas
        else:
            self.cost = 0



import base64
import time
from pathlib import Path
import aiofiles

import aiohttp
from nonebot import get_bot, on_command
from nonebot.adapters.onebot.v11 import GroupMessageEvent, Message, MessageSegment
from nonebot.log import logger
from nonebot.params import CommandArg
import re
import hashlib
from .config import config
from .requests import txt2img_body, header, htags, img2img_body
from .utils import is_contain_chinese,file_name_check
from .others.translation import translate
from .version import check_update
from .others.anlas import anlas_check, anlas_set, superusers
from .fifo import IMG2IMG, FIFO_IMG, TXT2IMG, FIFO_TXT
path = Path("data/novelai").resolve()
txt2img = on_command(".aidraw", aliases={"绘画", "咏唱","约稿","召唤"})

cd = {}
gennerating = False
limit_list = []


@txt2img.handle()
async def txt2img_handle(event: GroupMessageEvent, args: Message = CommandArg()):
    message_raw = args.extract_plain_text().replace("，", ",").split("-")
    user_id=str(event.user_id)
    count = 1
    # 以图生图预处理
    img_url = []
    for seg in event.message['image']:
        img_url.append(seg.data["url"])
    imgbytes: list[str] = []
    if img_url:
        if config.novelai_paid:
            async with aiohttp.ClientSession() as session:
                logger.info(f"正在获取图片")
                for i in img_url:
                    async with session.get(i) as resp:
                        imgbytes.append(await resp.read())
        else:
            await txt2img.finish(f"以图生图功能已禁用")
    if len(imgbytes) > config.novelai_oncemax:
        await txt2img.finish(f"最大只能同时生成{config.novelai_oncemax}张")

    # 功能开关
    match message_raw[0]:
        case "off":
            result = config.set_enable(event.group_id, False)
            logger.info(result)
            txt2img.finish(result)
        case "on":
            result = config.set_enable(event.group_id, True)
            logger.info(result)
            txt2img.finish(result)

    # 判断是否禁用，若没禁用，进入处理流程
    if event.group_id not in config.novelai_ban:

        # 判断cd
        nowtime = time.time()
        if (nowtime - cd.get(user_id, 0)) < config.novelai_cd:
            await txt2img.finish(f"你冲的太快啦，请休息一下吧")
        else:
            cd[user_id] = nowtime

        width,height = [512, 768]  # w*h
        tags = ""
        seed_raw = None
        nopre = False

        # 提取参数
        for i in message_raw:
            match i:
                case "square" | "s":
                    width,height = [640, 640]
                case "portrait" | "p":
                    width,height = [512, 768]
                case "landscape" | "l":
                    width,height = [768, 512]
                case "nopre" | "np":
                    nopre = True
                case _:
                    if i.isdigit():
                        seed_raw = int(i)
                    else:
                        tags += i

        if not tags:
            await txt2img.finish(f"请描述你想要生成的角色特征(使用英文Tag,代码内已包含优化TAG)")

        # 检测是否有18+词条
        if config.novelai_h:
            for i in htags:
                if i in tags.lower():
                    await txt2img.finish("H是不行的!")

        # 处理奇奇怪怪的输入
        tags = re.sub("\s", "", tags)
        tags = file_name_check(tags)

        # 生成种子
        seed = seed_raw or int(time.time())

        # 检测中文
        if is_contain_chinese(tags):
            tags_en = await translate(tags,"en")
            if tags_en==tags:
                txt2img.finish(f"检测到中文，翻译失败，生成终止，请联系BOT主查看后台")
            else:
                tags=tags_en
            logger.info(f"检测到中文，机翻结果为{tags}")
        if imgbytes:
            data_img = []
            for i in imgbytes:
                data_img.append(IMG2IMG(i))
            fifo = FIFO_IMG(user_id, tags, seed,
                            data_img, event.group_id)
            if fifo.cost > 0 and fifo.user_id not in superusers:
                anlascost = fifo.cost
                hasanlas = await anlas_check(fifo.user_id)
                if hasanlas>anlascost:
                    await wait_fifo(fifo,anlascost,hasanlas-anlascost)
                else:
                    await txt2img.finish(f"你的点数不足，你的剩余点数为{hasanlas}")
            else:
                await wait_fifo(fifo)
        else:
            data_txt = TXT2IMG(width,height)
            fifo = FIFO_TXT(user_id, tags, seed,
                            data_txt, event.group_id)
            await wait_fifo(fifo)


async def wait_fifo(fifo,anlascost=None,anlas=None):
    list_len = get_wait_num()
    has_wait = f"排队中，你的前面还有{list_len}人"
    no_wait = "请稍等，图片生成中"
    if anlas:
        has_wait += f"\n本次生成消耗点数{anlascost},你的剩余点数为{anlas}"
        no_wait += f"\n本次生成消耗点数{anlascost},你的剩余点数为{anlas}"
    if config.novelai_limit:
        await txt2img.send(has_wait if list_len > 0 else no_wait)
        limit_list.append(fifo)
        await run_txt2img()
    else:
        await txt2img.send(no_wait)
        await run_txt2img(fifo)


def get_wait_num():
    list_len = len(limit_list)
    if gennerating:
        list_len += 1
    return list_len


async def run_txt2img(fifo = None):
    global gennerating
    bot = get_bot()

    async def generate(fifo):

        logger.info(
            f"队列剩余{get_wait_num()}人 | 开始生成：{fifo.group_id},{fifo.user_id},{fifo.tags}")
        try:
            if isinstance(fifo,FIFO_IMG):
                im = await _run_img2img(fifo)
            else:
                im = await _run_txt2img(fifo)
        except:
            logger.exception("生成失败")
            im = "生成失败，请联系BOT主排查原因"
        else:
            logger.info(f"队列剩余{get_wait_num()}人 | 生成完毕：{fifo}")

        await bot.send_group_msg(
            message=MessageSegment.at(fifo.user_id) + im,
            group_id=fifo.group_id,
        )

    if fifo:
        await generate(fifo)

    if not gennerating:
        logger.info("队列开始")
        gennerating = True

        while len(limit_list) > 0:
            fifo = limit_list.pop(0)
            try:
                await generate(fifo)
            except:
                logger.exception("生成中断")

        gennerating = False
        logger.info("队列结束")
        await check_update()


async def _run_txt2img(fifo: FIFO_TXT):
    async with aiohttp.ClientSession(
        config.novelai_api_domain, headers=header
    ) as session:
        async with session.post(
            "/ai/generate-image", json=txt2img_body(fifo.seed, fifo.tags, fifo.data.width, fifo.data.height)
        ) as resp:
            if resp.status != 201:
                return f"生成失败，错误代码为{resp.status}"

            img = await resp.text()

        img_bytes = img.split("data:")[1]

        await save_img(fifo.seed, fifo.tags, img_bytes)
        return f"Seed: {fifo.seed}" + MessageSegment.image(f"base64://{img_bytes}")


async def _run_img2img(fifo: FIFO_IMG):
    img_byteslist = []
    async with aiohttp.ClientSession(
        config.novelai_api_domain, headers=header
    ) as session:
        for i in fifo.data:
            async with session.post(
                "/ai/generate-image", json=img2img_body(fifo.seed, fifo.tags, i.width, i.height, i.image)
            ) as resp:
                if resp.status != 201:
                    return f"生成失败，错误代码为{resp.status}"
                img = await resp.text()
                img_bytes = img.split("data:")[1]
            img_byteslist.append(img_bytes)
            await save_img(fifo.seed, fifo.tags, img_bytes)
    message = f"Seed: {fifo.seed}"
    for img_bytes in img_byteslist:
        message += MessageSegment.image(f"base64://{img_bytes}")
    if fifo.cost>0:
        await anlas_set(fifo.user_id,-fifo.cost)
    return message


async def save_img(seed, tags, img_bytes):
    if config.novelai_save_pic:
        path.mkdir(parents=True, exist_ok=True)
        img = base64.b64decode(img_bytes)
        hash = hashlib.md5(img).hexdigest()
        if len(tags) > 100:
            async with aiofiles.open(
                str(path/"output"/f"{seed}_{hash}_{tags[:100]}.png"), "wb"
            ) as f:
                await f.write(img)
        else:
            async with aiofiles.open(
                str(path/"output"/f"{seed}_{hash}_{tags}.png"), "wb"
            ) as f:
                await f.write(img)



from PIL import Image
import base64
from io import BytesIO

import urllib.parse
from .config import config
lowQuality = "lowres, bad anatomy, bad hands, text, error, missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, normal quality, jpeg artifacts, signature, watermark, username, blurry,"
htags = ["nsfw", "nude"]
basetag = "{masterpiece}, extremely detailed 8k wallpaper, best quality, an extremely delicate and beautiful,"
token = config.novelai_token
header = {
    "authorization": "Bearer " + token,
    ":authority": urllib.parse.urlparse(config.novelai_api_domain).hostname,
    ":path": "/ai/generate-image",
    "content-type": "application/json",
    "referer": config.novelai_site_domain,
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",
}


def txt2img_body(seed, input, width,height):
    if config.novelai_h:
        uc = lowQuality
    else:
        htagstr = ""
        for i in htags:
            htagstr = htagstr+i+","
        uc = lowQuality+htagstr
    return {
        "input": input + basetag + config.novelai_tag,
        "model": "safe-diffusion",
        "parameters": {
            "width": width,
            "height": height,
            "scale": 11,
            "sampler": "k_euler_ancestral",
            "qualityToggle": True,
            "steps": 28,
            "seed": seed,
            "n_samples": 1,
            "ucPreset": 0,
            "uc": uc,
        },
    }


def img2img_body(seed, input, width,height, image):
    if config.novelai_h:
        uc = lowQuality
    else:
        htagstr = ""
        for i in htags:
            htagstr = htagstr+i+","
        uc = lowQuality+htagstr
    return {
        "input": input + basetag + config.novelai_tag,
        "model": "safe-diffusion",
        "parameters": {
            "width": width,
            "image": image,
            "noise": 0.2,
            "strength": 0.7,
            "height": height,
            "qualityToggle": True,
            "scale": 11,
            "sampler": "k_euler_ancestral",
            "steps": 50,
            "seed": seed,
            "n_samples": 1,
            "ucPreset": 0,
            "uc": uc,
        },
    }



def is_contain_chinese(check_str:str):
    for ch in check_str:
        if u'\u4e00' <= ch <= u'\u9fff':
            return True
    return False

import aiohttp


def file_name_check(input:str):
    errorcode=["?","<",">","\\","/","*","|",":"]
    for i in errorcode:
        input.replace(i,"")
    return input

import re
async def check_last_version(package:str,version:str):
    pypiversion=""
    versionlist=version.split(".")
    async with aiohttp.ClientSession() as session:
        async with session.get("https://pypi.org/simple/"+package) as resp:
            text=await resp.text()
            pattern=re.compile("-(\d.*?).tar.gz")
            pypiversion=re.findall(pattern,text)[-1]
    pypilist:list[str]=pypiversion.split(".")
    for i in range(len(versionlist)):
        if int(pypilist[i])>int(versionlist[i]):
            return pypiversion
        elif int(pypilist[i])<int(versionlist[i]):
            return None
    return None

async def sendtosuperuser(message):
    # 将消息发送给superuser
    from nonebot import get_bot,get_driver
    import asyncio
    superusers=get_driver().config.superusers
    bot=get_bot()
    for superuser in superusers:
        await bot.call_api('send_msg', **{
            'message': message,
            'user_id': superuser,
        })
        await asyncio.sleep(5)
